
/*	AY-3-8912 sound chip emulation

***	Copyright:	1995-1999 Gnter Woigk
  				mailto:kio@myokay.net

 	This program is distributed in the hope that it will be useful,
 	but WITHOUT ANY WARRANTY; without even the implied warranty of
 	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  

 	Permission to use, copy, modify, distribute, and sell this software and
 	its documentation for any purpose is hereby granted without fee, provided
 	that the above copyright notice appear in all copies and that both that
 	copyright notice and this permission notice appear in supporting
 	documentation, and that the name of the copyright holder not be used
 	in advertising or publicity pertaining to distribution of the software
 	without specific, written prior permission.  The copyright holder makes no
 	representations about the suitability of this software for any purpose.
 	It is provided "as is" without express or implied warranty.

 	THE COPYRIGHT HOLDER DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS SOFTWARE,
 	INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS, IN NO
 	EVENT SHALL THE COPYRIGHT HOLDER BE LIABLE FOR ANY SPECIAL, INDIRECT OR
 	CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE,
 	DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER
 	TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR
 	PERFORMANCE OF THIS SOFTWARE.

	24.Mar.95:	First revision					KIO !
	05.Jun.95:	Combined ear and mic sound output		KIO !
	29.Nov.95:	AY-3-8912 emulation				KIO !
	05.Jan.99:	port to Linux					kio
	09.jan.99:	ay emu fixed, perfect sound sampling		kio
*/


/* ==================================================================
		Hardware independent part
================================================================== */


#include <sys/ioctl.h>
#include <unistd.h>
#include <fcntl.h>
#include <sys/soundcard.h>


#include "h/options.h"
#include "kio.h"
#include "z80/z80.h"
#include "h/audio.h"
#include "h/ay.h"
#include "h/main.h"


#ifndef LOG
#define LOG 1
#endif
#ifndef SAFE
#define SAFE 2
#endif

#define	SPEEDPATCHES	on

#define	SOUNDLOGSIZE	2000			// size of output log buffer

#define	VOLUME		(((Sample)-1)/6)	// volume per channel and ear+mic
#define	QUIET		(1+VOLUME*3)		// should be approx. 0x80

#define	low		false			// signal levels
#define	high		true


/* ===============================================================
		amplifier stage of a channel
=============================================================== */


/* ----	logarithmic volume table ---------------------------------
*/
static Sample logvol[16];		// logarithmic volume table


/* ----	de/select envelope for this channel --------------------
*/
inline void Channel::UseEnvelope ( bool env )
{
	use_envelope = env;
	volume = env ? envelopevolume : soundvolume;
	if (amp_in) output = volume;
}


/* ----	set volume and select envelope --------------------------
	in: value written to the channel's volume register
	each channel has it's own amplifier.
	changing the volume takes effect immediately
	bits 0-3 define the channel's volume, which is stepped logarithmically
	and bit 4 overrides bits 0-3 selecting the envelope generator's volume
*/
void Channel::SetVolume ( Char n )
{
	soundvolume = logvol[n&0x0f];
	UseEnvelope(n&0x10);
}


/* ----	set envelope volume -------------------------------------
	in: new sample value of envelope output
	immediately sets new volume if envelope is enabled
*/
void Channel::SetEnvelopeSample ( Sample n )
{
	envelopevolume = n;
	if (use_envelope) { volume = n; if (amp_in) output = n; }
}


/* ----	set envelope volume -------------------------------------
	in: new volume of envelope output
	immediately sets new volume if envelope is enabled
*/
inline void Channel::SetEnvelopeVolume ( Char n )
{
	XXXTRAP(n>=16);
	SetEnvelopeSample(logvol[n]);
}


/* ----	change state of binary sound input ----------------------
*/
inline void Channel::AmpInput ( bool in )
{
	amp_in = in;
	output = in ? volume : 0;
}


/* ----	get output of amplifier stage ---------------------------
*/
inline Sample Channel::GetOutput ( )
{
	return output;
}


/* ===============================================================
		mixer stage of a channel
=============================================================== */


/* ----	enable this channel -------------------------------------
	this is an additional enable, which is not part of the ay chip.
	it is provided for debugging etc.
*/
void Channel::EnableChannel ( bool f )
{
	channel_enable = f;
	AmpInput ( channel_enable && sound_out && noise_out );
}


/* ----	enable sound generation ----------------------------------
	in: !(bit_from_mixer_control_register)

  note:	in a real ay chip, sound output to the amplifier is 
	calculated from the mixer inputs like this:
     =>	output = (disable_sound||sound_input) && (disable_noise||noise_input)
	in this emulation we use the equivalent calculation:
     =>	output = (!enable_sound||sound_input) && (!enable_noise||noise_input) && channel_enable
*/
void Channel::EnableSound ( bool f )
{
	sound_enable = f;
	sound_out = sound_in || !f;
	AmpInput ( channel_enable && sound_out && noise_out );
}


/* ----	enable noise generation ----------------------------------
	in: !(bit_from_mixer_control_register)
*/
void Channel::EnableNoise ( bool f )
{
	noise_enable = f;
	noise_out = noise_in || !f;
	AmpInput ( channel_enable && sound_out && noise_out );
}


/* ----	change input from noise generator -------------------------
*/
inline void Channel::MixNoiseInput ( bool in )
{
	noise_in = in;

	if (noise_enable)
	{
		noise_out = in;
		AmpInput ( channel_enable && sound_out && noise_out );
	}
}


/* ----	change input from sound generator -------------------------
*/
inline void Channel::MixSoundInput ( bool in )
{
	sound_in = in;

	if (sound_enable)
	{
		sound_out = in;
		AmpInput ( channel_enable && sound_out && noise_out );
	}
}


/* ===============================================================
		sound generator stage of a channel
=============================================================== */


/* ----	reset the whole channel ----------------------------------
	reset sound generator, mixer and amplifier
	does not change the overall channel_enable state !
	set divider for cpu cycles -> ay internal cycles
*/
void Channel::Reset ( long div )
{
// reset counter stage
	sound_in = low;
	fine     = 0xff;
	coarse   = 0x0f;
	divider  = div;
	reload   = 0x0fff*div;
	when     = reload;

// reset mixer stage
	EnableSound(false);
	EnableNoise(false);
	// EnableChannel(channel_enable);	does not change enable state !

// reset amplifier stage
	SetVolume(0);		// volume=0 & envelope off
	SetEnvelopeSample(0);
}


/* ----	creator ---------------------------------------------------
	disable channel
	reset all stages
*/
Channel::Channel ( long div )
{
	EnableChannel(false);
	Reset(div);
}


/* ----	change a channel's pitch ----------------------------------
	the ay input clock is first divided by 16. then this clock is fed into a counter
	which's output is the mixer's "Soundin" state for this channel.
	programming a pitch of 0 seems to have the same effect as a pitch of 1
	each counter has 12 significant bits.

	it is not known when a new value takes effect:
	repeatedly programming the same value has no effect on the counter
	it may be a count down which restores the reload value when it reaches/underflows 0
	it may be a count up which is reset to 0 when it reaches/overlows the reload value

	the current implementation assumes that it is a count down and that the
	new value takes effect when the counter is reload
*/
void Channel::SetFinePeriod ( Char n )
{
	fine = n;
	reload = (coarse*256+n)*divider; 
	if (!reload) reload=divider;
}


void Channel::SetCoarsePeriod ( Char n )
{
	n &= 0x0f;
	coarse = n;
	reload = (n*256+fine)*divider;
	if (!reload) reload=divider;
}


void Channel::Trigger ( )
{
	MixSoundInput ( !sound_in );
	when += reload;
}


/* ===============================================================
			noise generator
=============================================================== */


// white noise is generated using this run length table
static Char ndur[2048];		// pseudo-random reload value table


/* ----	reset the noise generator --------------------------------
*/
void Noise::Reset ( long div )
{	bool s = high;

	divider = div;
	reload  = 0x1f * div;
	when    = reload;
	index   = 0;
	state   = s;
	A->MixNoiseInput(s);
	B->MixNoiseInput(s);
	B->MixNoiseInput(s);
}


/* ----	creator ---------------------------------------------------
	since the noise generator actively propagates it's output state 
	to the sound channels, it needs to know where they are
*/
Noise::Noise ( long div, Channel* a, Channel* b, Channel* c)
{
	A=a;
	B=b;
	C=c;
	Reset(div);
}


/* ----	set noise counter period -----------------------------------
	in: Value written to noise period register

	there is 1 noise generator. it's output may be mixed to any channel.
	the ay input clock is first divided by 16. then this clock is fed into a counter
	which's output is the mixer's "Noisein" state for every channel.
	programming a pitch of 0 seems to have the same effect as a pitch of 1
	this counter has 5 significant bits.

	the algorithm for creating the noise is not known. it seems, that every time
	the counter expires it's output decides on a pseudo random base whether 
	to keep it's state or toggle. 

	the implementation of counters seems to be very similar for channels and noise.
	so the following is the same as for channels:

	it is not known when a new value takes effect:
	repeatedly programming the same value has probably no effect on the counter
	it may be a count down which restores the reload value when it reaches/underflows 0
	it may be a count up which is reset to 0 when it reaches/overlows the reload value

	the current implementation assumes that it is a count down and that the
	new value takes effect when the counter is reload
*/
inline void Noise::SetPeriod ( Char n )
{
	n&=0x1f; n+=!n;
	reload = n*divider;
}


/* ----	Trigger noise counter --------------------------------------
	Trigger() must be called when according to When() the counter expired
	if index==-1 then noise is disabled!
*/
void Noise::Trigger ( )
{
	if (index<=0)	// wrap in ndur[] ?
	{
#if SPEEDPATCHES
		if (index==-1)		// noise disabled ?
		{ 	when += 0x7fffffff/reload*reload; 
			return; 	// return without toggle
		}
#endif
		index = NELEM(ndur);
	}

	when += reload*ndur[--index];
	bool s = state = !state;
	A->MixNoiseInput(s);
	B->MixNoiseInput(s);
	C->MixNoiseInput(s);
}


/* ----	en/disable noise output --------------------------------------
	noise output is re-enabled with Enable(true)
	noise output is disabled with Enable(false)
	disabled noise generator is flagged with index=-1
	enabled noise starts at random position in ndur[] with random delay
	disabled nois generator propagates high output to sound channel mixers
*/
#if SPEEDPATCHES
void Noise::Enable ( long now, bool f )
{
	if (f)	
	{	if (index==-1)		// enable
		{
			XXLog("noise ON");
			index = Random(NELEM(ndur)-1);	// random start position
			when = now - (Long)now%reload + reload*ndur[index];	// start offset
		}
	}
	else if (index!=-1)		// disable
	{
		XXLog("noise OFF");
		index = -1;		// marker: noise disabled
		state = high;		// => enabling noise in mixer has no effect
		A->MixNoiseInput(high);	// propagate output to channels' mixer
		B->MixNoiseInput(high);
		C->MixNoiseInput(high);
	}
}
#endif


/* ===============================================================
			envelope generator
=============================================================== */


/* ----	internal representation of the sound chip's state -------------------------

	envelope generator:
	there is 1 envelope generator. 
	envelope shape is controlled by attack, invert and hold.
	the 'repeat' bit (D3 of reg. 13) is not used and always treated as set:
	therefore %00xx is converted to %1001 and %01xx is converted to %1111 
	which give the same results.

	D3 = repeat
	D2 = attack
	D1 = toggle
	D0 = hold

	1111, 01xx:	/_____	 repeat &  attack &  toggle &  hold	||  !repeat & attack
	1110:		/\/\/\	 repeat &  attack &  toggle & !hold
	1101:		/~~~~~	 repeat &  attack & !toggle &  hold
	1100:		//////	 repeat &  attack & !toggle & !hold
	1011:		\~~~~~	 repeat & !attack &  toggle &  hold
	1010:		\/\/\/	 repeat & !attack &  toggle & !hold
	1001, 00xx:	\_____	 repeat & !attack & !toggle &  hold	||  !repeat & !attack
	1000:		\\\\\\	 repeat & !attack & !toggle & !hold

	the envelope generator is running free and only retriggered 
	when a new envelope is selected.
*/

// can the additional 4 bit predivider of the envelope generator be reset ?
// probably true, so set it to true
#define	E_PREDIV_RESET	false


/* ----	reset the envelope generator -------------------------------------------
*/
void Envelope::Reset ( long div )
{
	divider = div;//*16;
	reload  = 0x3fffffff/divider*divider;
	when    = reload;
	fine    = coarse = 0xff;
	index   = 0;
	repeat  = invert = off;
	direction = 0;

	A->SetEnvelopeSample(0);
	B->SetEnvelopeSample(0);
	C->SetEnvelopeSample(0);
}


/* ----	creator ---------------------------------------------------------------
	since the envelope generator actively propagates it's output state 
	to the sound channels, it needs to know where they are
*/
Envelope::Envelope ( long div, Channel* a, Channel* b, Channel* c )
{
	A=a;
	B=b;
	C=c;
	Reset(div);
}


/* ----	set counter low byte -------------------------------------------------
	in: byte written to envelope fine period register
*/
void Envelope::SetFinePeriod ( Char n )
{
	fine = n;
	reload = ((long)coarse*256+n)*divider;
	if (!reload) reload=divider;
}


/* ----	set counter high byte -------------------------------------------------
	in: byte written to envelope coarse period register
*/
void Envelope::SetCoarsePeriod ( Char n )
{
	coarse = n;
	reload = ((long)n*256+fine)*divider;
	if (!reload) reload=divider;
}


/* ----	select new envelope shape -------------------------------------------
	in: byte written to the envelope control register
	this automatically restarts the envelope generator
*/
void Envelope::SetShape ( long now, Char c )
{
	if (!(c&8)) c = c&4 ? 0x0f : 0x09;

	index	  = BIT(2,c) ? 0 : 15;

	direction = BIT(2,c) ? 1 : -1;
	invert    = BIT(1,c); 
	repeat    = !BIT(0,c);

#if E_PREDIV_RESET
	when	  = (Long)now-(Long)now%(divider/16) + reload;
#else
	when	  = (Long)now-(Long)now%divider + reload;
#endif

	Sample v = logvol[index];
	A->SetEnvelopeSample(v);
	B->SetEnvelopeSample(v);
	C->SetEnvelopeSample(v);
}


/* ---- trigger the envelope counter ---------------------------------------
	Trigger() must be called when according to When() the counter expired
*/
void Envelope::Trigger ( )
{
	index += direction;
	when  += reload;
	
	if (index & 0xF0)		// ramp finished
	{
		if (repeat)
		{	if (invert) { index = ~index; direction = -direction; }
		}
		else
		{	direction=0; if (!invert) index = ~index;
			when += 0x3fffffff/divider*divider -reload;
		}
		index  &= 0x0F;
	}

	Sample v = logvol[index];
	A->SetEnvelopeSample(v);
	B->SetEnvelopeSample(v);
	C->SetEnvelopeSample(v);
}


/* ===============================================================
			the beeper
=============================================================== */


inline void Beeper::Reset ( )
{
	state   = EAR_MASK*0 + MIC_MASK*1;
	sample  = QUIET;
	
}


inline Beeper::Beeper ( )
{
	enabled = 0;
	Reset();
}


/* ----	change state of ear and mic output ------------
cssfaq:	The lowest three bits specify the border colour; 
	a zero in bit 3 activates the MIC output, 
	whilst a one in bit 4 activates the EAR output and the internal speaker. 
	However, the EAR and MIC sockets are connected only by resistors, 
	so activating one activates the other; 
	the EAR is generally used for output as it produces a louder sound. 
	The upper two bits are unused.
*/
inline Sample Beeper::SetInput ( Char n )
{
	state = n;
	n &= enabled;
	return sample =
		( n&EAR_MASK ? VOLUME*6/4 : 0 ) + 	// high => ear out (louder) + internal speaker
		( n&MIC_MASK ? 0 : VOLUME*2/4 );	// low  => mic out (softer)
}


inline Sample Beeper::GetOutput ( )
{
	return sample; 
}


void Beeper::Enable ( Char f )
{
	enabled = f & (EAR_MASK+MIC_MASK);
	SetInput(state);
}


void Beeper::EnableEar ( bool f )
{
	Enable ( (enabled&MIC_MASK) + f*EAR_MASK );
}


void Beeper::EnableMic ( bool f )
{
	Enable ( (enabled&EAR_MASK) + f*MIC_MASK );
}


/* ===============================================================
			ay sound chip
=============================================================== */


/* ----	set ay register with all effects -------------------------
*/
void ay::SetRegister ( long now, int nr, Char c )
{
	XXLogLine("ay::SetRegister(%i,%i,%i)", now,nr,(int)c);

	switch (nr)
	{
	case 0:  A.SetFinePeriod(c);	return;
	case 1:  A.SetCoarsePeriod(c);	return;
	case 2:  B.SetFinePeriod(c);	return;
	case 3:  B.SetCoarsePeriod(c);	return;
	case 4:  C.SetFinePeriod(c);	return;
	case 5:  C.SetCoarsePeriod(c);	return;
	case 6:  N.SetPeriod(c);	return;
	case 7:  c = ~c;
#if SPEEDPATCHES
		 N.Enable(now, c&0x38);
#endif
		 A.EnableSound(c&1);
		 B.EnableSound(c&2);
		 C.EnableSound(c&4);
		 A.EnableNoise(c&8);
		 B.EnableNoise(c&16);
		 C.EnableNoise(c&32);	return;
	case 8:  A.SetVolume(c);	return;
	case 9:  B.SetVolume(c);	return;
	case 10: C.SetVolume(c);	return;
	case 11: E.SetFinePeriod(c);	return;
	case 12: E.SetCoarsePeriod(c);	return;
	case 13: E.SetShape(now,c);	return;

// the i/o ports are not yet implemented:
	case 14: return; // 1st i/o port: not implemented
	case 15: return; // 2nd i/o port: not implemented

// the beeper is handled by the ay emulation too:
	case 16: beeper.SetInput(c);	return;

// nop fall-through
	case 17: return;

	default: IERR();
	}
}


/* ----	Reset sound chip and beeper -----------------------
*/
void ay::Reset ( long cpuclk, long ayclk )
{
	XLogLine ( ">ay::Reset(%i,%i)", cpuclk, ayclk );
	
// init clocks
	ayclock   = ayclk;		// ay clock
	cpuclock  = cpuclk;		// cpu clock
	aycycles  = cpuclk*8/ayclk;	// full cpu cycles per half ay clock cycle after predivision by 16
	dspcycles = cpuclk/dsp.Clock();	// cpu cycles per dsp sample (( only valid if dsp.On() ))

// reset all externally writable registers
	const Char m[] = 
	{ 	0xff,0x0f,0xff,0x0f,0xff,0x0f, 	// channel A,B,C period
		0x1f, 				// noise period
		0x3f, 				// mixer control
		0x00,0x00,0x00, 		// channel A,B,C volume
		0xff,0xff,0x00, 		// envelope period and shape
		0xff,0xff, 			// i/o ports
		0xff				// beeper port
	};
	MemCopy ( m, AYreg, 17 );
	AYregnr   = 7; 			

// reset generators
	A.Reset(aycycles);
	B.Reset(aycycles);
	C.Reset(aycycles);
	N.Reset(aycycles);
	E.Reset(aycycles);
	beeper.Reset();

	XLogExit();
}


/* ----	standard creator -----------------------------------
	all channels are disabled and must be enabled with EnableXXX()
	sound output is disabled and must be enabled with Resume()
*/
ay::ay ( long cpuclk, long ayclk )
:	N(8,&A,&B,&C),
	E(8,&A,&B,&C),
	A(8),
	B(8),
	C(8)
{
	XLogLine(">ay::ay()");

// calculate random duration table for white noise
	for ( int i=NELEM(ndur); i--; )
	{
		Char n;
		for ( n=1;n<16;n++ ) if (random()&1) break;
		ndur[i] = n;
	};

// calculate logarithmic volume table
	long n = VOLUME*100;
	for ( int i=16; i--; )
	{	logvol[i] = n/100;
		n = n * 82 / 100;
	};
	logvol[0] = 0;

// init valid bit masks
	const Char m[] = 
	{ 	0xff,0x0f,0xff,0x0f,0xff,0x0f, 	// channel A,B,C period
		0x1f, 				// noise period
		0x3f, 				// mixer control
		0x1f,0x1f,0x1f, 		// channel A,B,C volume
		0xff,0xff,0x0f, 		// envelope period and shape
		0xff,0xff 			// i/o ports
	};
	MemCopy ( m, AYregMask, 16 );

// reset internal state
	enabled = 0;
	Reset ( cpuclk, ayclk );

	XLogExit();
}


/* ----	destructor ----------------------------------------
*/
ay::~ay ( )
{
	XLogLine(">ay::~ay()");
	Suspend();
	XLogExit();
}


/* ----	Suspend sound generation --------------------------
	stop sound immediately
	dispose off all channels and memory
	internal state of the ay chip remains valid!
	ay::SetRegister() etc. still update the registers
	so that sound generation may be restarted any time
*/
void ay::Suspend ( )
{
	XLogLine ( ">ay::Suspend()" );

	if (dsp.On()) dsp.Close();

	delete[] sndlog;
	sndlog = sndlogptr = sndlogend = 0;

	XLogExit();
}


/* ----	Resume sound generation ------------------------------
	This function completely initializes the sound module
*/
void ay::Resume ( long now )
{
	XLogLine(">ay::Resume(%i)", now);

	if ( !dsp.On() && dsp.Open()==ok )
	{
		OMEM ( ( sndlog = sndlogptr = new aylog[SOUNDLOGSIZE+1] ) );
		sndlogend = sndlog+SOUNDLOGSIZE;

	// how many clock cycles fit in a sample ?
		dspcycles = cpuclock/dsp.Clock();

	// restore internal state of ay chip and beeper
		for ( uint i=0; i<NELEM(AYreg); i++ ) SetRegister(now,i,AYreg[i]);

	// internal state of subsample collector
		oldcycle  = now;
		subcycle  = now;
		subsample = 0;
		oldsample = QUIET;	// QUIET+a.output+b.output+c.output-beeper.output
	}

	XLogExit();
}


/* ----	Enable this or that channel ------------------------------
	this function does not start/stop sound output!
	sound output may still be on even if all channels are disabled!
*/
void ay::Enable ( int enab )
{
	XLogLine(">ay::Enable(%i)", enab);

	A.EnableChannel(enab&1);
	B.EnableChannel(enab&2);
	C.EnableChannel(enab&4);
	beeper.Enable(enab);

	enabled = enab;
	
	XLogExit();
}

void ay::EnableAY 	 ( bool f )	{ Enable ( (enabled&~7) + f*7 ); }
void ay::EnableChannelA  ( bool f )	{ Enable ( (enabled&~1) + f*1 ); }
void ay::EnableChannelB  ( bool f )	{ Enable ( (enabled&~2) + f*2 ); }
void ay::EnableChannelC  ( bool f )	{ Enable ( (enabled&~4) + f*4 ); }
void ay::EnableBeeper    ( bool f )	{ Enable ( (enabled&~(EAR_MASK+MIC_MASK)) + f*(EAR_MASK+MIC_MASK) ); }
void ay::EnableEarOutput ( bool f )	{ Enable ( (enabled&~EAR_MASK) + f*EAR_MASK ); }
void ay::EnableMicOutput ( bool f )	{ Enable ( (enabled&~MIC_MASK) + f*MIC_MASK ); }



/* ----	output samples to the dsp ---------------------------
*/
void ay::WriteSample ( long now, Sample sample )
{
	XXXLogLine ( ">ay::WriteSample(%i,%i)",now,sample);

	if ( (now-subcycle)>=dspcycles )
	{
		dsp.Write ( (subsample+oldsample*(subcycle+dspcycles-oldcycle))/dspcycles );
		subcycle += dspcycles;			

		while ((now-subcycle)>=dspcycles)	
		{						
			dsp.Write(oldsample);			
			subcycle += dspcycles;
		}						

		oldcycle = subcycle;
		subsample = 0;
	}

	subsample += (now-oldcycle)*oldsample;
	oldcycle  = now;
	oldsample = sample;

	XXXLogExit();
}


/* ----	Add one frame of sampled sound ----------------------------------------------------
*/
void ay::Play ( long now )
{
	XXLogLine ( ">ay::Play(%i) -> %i Outputs, %i T cycles", now, sndlogptr-sndlog, now-oldcycle );
	XTRAP (!sndlog);
	XTRAP (!dsp.On());
	XXLog("oldsample=%i", this->oldsample);

// load old state:
//	int  oldsample	= this->oldsample;
//	long oldcycle	= this->oldcycle;
//	int  subsample	= this->subsample;
//	long subcycle	= this->subcycle;

// ---- DOIT ----

	if ( enabled&7 )	// channel A, B or C enabled ?
	{
	// ----------> sample sound: ear/mic/AY

		XXLog("ay+beeper");

	// end-of-list marker
		sndlogptr->reg=17;
		sndlogptr->cycle=now;

		for ( aylog* q=sndlog; q<=sndlogptr; )
		{

		// who is next ?
			long now;
			int who;
			{ who=1; now=N.When(); }	// assume: noise is next
			if ( E.When()-now<0 )	
			{ who=2; now=E.When(); }	// envelope is next
			if ( A.When()-now<0 )
			{ who=3; now=A.When(); }	// channel A is next
			if ( B.When()-now<0 )
			{ who=4; now=B.When(); }	// channel B is next
			if ( C.When()-now<0 )
			{ who=5; now=C.When(); }	// channel C is next
			if ( q->cycle-now<0 ) 
			{ who=6; now=q->cycle; }	// new command is next

			XXXLog("-%i-",who);

		// handle next:
			switch (who)
			{
			case 1: N.Trigger(); break;
			case 2: E.Trigger(); break;
			case 3: A.Trigger(); break;
			case 4: B.Trigger(); break;
			case 5: C.Trigger(); break;

			default:// next command
				SetRegister(now,q->reg,q->value);
				q++;
				break;
			}

			WriteSample ( now, QUIET + A.GetOutput()+B.GetOutput()+C.GetOutput()-beeper.GetOutput() );
		}
	}

	else 
	{
		if ( enabled&(EAR_MASK+MIC_MASK) )
		{
		// -------------> sample sound: ear/mic only
			XXLog("beeper only");
	
			for ( aylog* q=sndlog; q<sndlogptr; q++)
			{	
				if (q->reg==16) WriteSample ( q->cycle, QUIET -beeper.SetInput(q->value) );
			};
		}
	
		else	XXLog("silence");
	
		WriteSample ( now, QUIET -beeper.GetOutput() );
	}
		
// store current state
//	this->oldsample	= oldsample;	
//	this->oldcycle	= oldcycle;
//	this->subsample	= subsample;
//	this->subcycle	= subcycle;

// clear sound log
	sndlogptr	= sndlog;

	XXLogExit();
}







