/*
 * Copyright (c) Des Herriott 1993, 1994
 * Copyright (c) Gnter Woigk 1999
 *
 * Permission to use, copy, modify, distribute, and sell this software and its
 * documentation for any purpose is hereby granted without fee, provided that
 * the above copyright notice appear in all copies and that both that
 * copyright notice and this permission notice appear in supporting
 * documentation, and that the name of the copyright holder not be used in
 * advertising or publicity pertaining to distribution of the software without
 * specific, written prior permission.  The copyright holder makes no
 * representations about the suitability of this software for any purpose.  It
 * is provided "as is" without express or implied warranty.
 *
 * THE COPYRIGHT HOLDER DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS SOFTWARE,
 * INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS, IN NO
 * EVENT SHALL THE COPYRIGHT HOLDER BE LIABLE FOR ANY SPECIAL, INDIRECT OR
 * CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE,
 * DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER
 * TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR
 * PERFORMANCE OF THIS SOFTWARE.
 *
 * Author:	Des Herriott
 *			Gnter Woigk
 */

/*
 * Source file util.c - contain various common utility functions
 */



#include 	<sys/types.h>
#include 	<sys/time.h>
#include 	<sys/stat.h>
#include 	<sys/mman.h>
#include 	<stdio.h>
#include 	<stdarg.h>
#include 	<stdlib.h>
#include 	<string.h>
# include 	<strings.h>
#include 	<ctype.h>
#include 	<unistd.h>
#include 	<fcntl.h>
#include	<assert.h>

#include 	"h/options.h"
#include 	"h/main.h"
#include 	"h/util.h"

#ifndef 	LOG
#define 	LOG 2
#endif
#ifndef 	SAFE
#define 	SAFE 1
#endif


/* -------------------------------------------------------
	quit handling
	Author: Des Herriott
	Adaptions G. Woigk
------------------------------------------------------- */


struct on_quit_rec 
{
	void		(*f)(void);	// routine to call before quitting
	on_quit_rec	*next;		// linked list
};


static on_quit_rec	*on_quit_list	= NULL;


/* ----	add routine to call before quitting ----------------------------
*/
void OnQuit ( void (*fn)(void) )
{
	XXLogLine ( ">OnQuit(*fn)" );

	on_quit_rec *p;

	OMEM ( p = (struct on_quit_rec *)malloc(sizeof (struct on_quit_rec)) );
	p->f = fn;
	p->next = on_quit_list;
	on_quit_list = p;

	XXLogExit();
}


/* ----	call all registered quit routines and exit -------------------------
*/
void Quit ( int status )
{
	struct on_quit_rec *p = on_quit_list;

	XXLogLine ( ">Quit(%i)", status );

	while (p)
	{
		p->f();
		p = p->next;
	}

	XXLogExit();

	exit(status);
}



/* -------------------------------------------------------
	Implementations of strcasecmp and strncasecmp
	if you don't have them 
	Author: Des Herriott
------------------------------------------------------- */

#ifdef NEED_STRCASECMP
int strcasecmp(char *s1, char *s2)
{
	char c1, c2;

	for ( ; *s1 && *s2; s1++, s2++) {
		c1 = isupper(*s1) ? tolower(*s1) : *s1;
		c2 = isupper(*s2) ? tolower(*s2) : *s2;
		if (c1 != c2) return (c1 - c2);
	}
	if (*s1 || *s2) return (*s1 - *s2);
	return 0;
}

int strncasecmp(char *s1, char *s2, int n)
{
	int i;
	char c1, c2;

	for (i = 0; *s1 && *s2 && i < n; s1++, s2++, i++) {
		c1 = isupper(*s1) ? tolower(*s1) : *s1;
		c2 = isupper(*s2) ? tolower(*s2) : *s2;
		if (c1 != c2) return (c1 - c2);
	}
	if ((i < n) && (*s1 || *s2)) return (*s1 - *s2);
	return 0;
}
#endif




/* -----------------------------------------------------------
	file utilities 
----------------------------------------------------------- */


/* ----	split file name from path ----------------------------
		Gnter Woigk
*/
const char* BaseName ( char* filename )
{
	char *s=0;
	return !filename ? "NULL" : s=strrchr(filename,'/') ? s+1 : filename;
}


/*
 * Decide whether file descriptor <fd> is a normal file, a tty, or a pipe.
 * This affects the semantics of reads and writes via RS232.
 * Author: Des Herriott
 */
int classify_descriptor(int fd)
{	int n;

	XXLogLine(">classify_descriptor(%i)", fd);

	struct stat buf;

	n=FD_TTY; if (isatty(fd)) goto x;

	if (!fstat(fd, &buf))
	{
		n=FD_FILE; if (S_ISREG(buf.st_mode))  goto x;
		n=FD_PIPE; if (S_ISFIFO(buf.st_mode)) goto x;
	}

	LogLine ( "can't classify fd %d", fd );
	n=FD_UNKNOWN;

x:	XXLogExit();
	return n;
}


/* Open a file, checking each element in the colon-separated resource
 * 'lib_dir', unless the filename is an absolute path or relative to the
 * current or parent directory.  Behaves like fopen() in other respects.
 * Author: Gnter Woigk
 * based on open_xzx_file(...) from Des Herriott
 */
FILE* Fopen ( char *name, const char *mode )
{	FILE *fp = NULL;
	char *path, *pathend, *fullname;
	char pathlist[] = ":" "./:" LIBDIR ":";

	if (!name) name="";

	XXLogLine ( ">Fopen(\"%s\",\"%s\")", name, mode );

	if (name[0]==0) goto x;

	OMEM ( fullname = new char[strlen(name)+strlen(LIBDIR)+1+1] );

	for ( path=pathlist; (pathend=strchr(path,':')); path=pathend+1 )
	{
		*pathend=0;
		sprintf ( fullname, "%s/%s", path, name );

		fp = fopen ( fullname, mode );
	 	if (fp) break;
	}

	XXLog ( " => \"%s\": %s", fullname, fp?"ok":"not found" );
	
	delete[] fullname;

x:	XXLogExit();

	return fp;
}































