
/*
 * Copyright (c) Des Herriott 1993, 1994
 * Copyright (c) Gnter Woigk 1999
 *
 * Permission to use, copy, modify, distribute, and sell this software and its
 * documentation for any purpose is hereby granted without fee, provided that
 * the above copyright notice appear in all copies and that both that
 * copyright notice and this permission notice appear in supporting
 * documentation, and that the name of the copyright holder not be used in
 * advertising or publicity pertaining to distribution of the software without
 * specific, written prior permission.  The copyright holder makes no
 * representations about the suitability of this software for any purpose.  It
 * is provided "as is" without express or implied warranty.
 *
 * THE COPYRIGHT HOLDER DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS SOFTWARE,
 * INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS, IN NO
 * EVENT SHALL THE COPYRIGHT HOLDER BE LIABLE FOR ANY SPECIAL, INDIRECT OR
 * CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE,
 * DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER
 * TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR
 * PERFORMANCE OF THIS SOFTWARE.
 *
 * Author: Des Herriott
 * (and Thomas Ahn Kjaer for the banked port handling)
 * Gnter Woigk
 */



#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include <sys/ipc.h>
#include <sys/shm.h>
#include <X11/extensions/XShm.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <assert.h>


#include "options.h"
#include "screen.h"
#include "window.h"
#include "util.h"
#include "main.h"
#include "keyboard.h"
#include "zxsp.xbm"


#ifndef LOG
#define LOG 1
#endif
#ifndef SAFE
#define SAFE 2
#endif



/* ---------------------------------------------------------
		Display of Screen contents
		Hardware dependent part
--------------------------------------------------------- */


/* ----	close and destroy window --------------------------
*/
Win::~Win ( )
{
	XXLogLine ( ">Win::~Win()" );

	DestroyImage();
	DestroyColors();
	XCloseDisplay(display);

	XXLogExit();
}


/* ----	create and open a window -----------------------------
*/
Win::Win ( int argc, char* argv[] )
{	int i,j;
	bool f;

	XLogLine ( ">Win::Win(%i,argv[])", argc );

	MemClear(this,sizeof(Win));


// get environment 
	display = XOpenDisplay(NULL);	// default display on POSIX
	if (!display) { Log ( "couldn't connect to X server"); done=true; goto x; }

	XLog("bitmapBitOrder = %s", BitmapBitOrder(display) == LSBFirst ? "LSB first" : "MSB first");
	XLog("imageByteOrder = %s", ImageByteOrder(display) == LSBFirst ? "LSB first" : "MSB first");

	screen = DefaultScreenOfDisplay(display);
	root   = RootWindowOfScreen(screen);
	white  = WhitePixelOfScreen(screen);
	black  = BlackPixelOfScreen(screen);
	XXXLog("w1");


// search for supported visual & pixel size
	{
	static int  visuals[] = { PseudoColor,  TrueColor,  DirectColor,  StaticColor,  StaticGray,  GrayScale };// supported visual types in preferences order
	static ptr  visinam[] = {"PseudoColor","TrueColor","DirectColor","StaticColor","StaticGray","GrayScale"};// their names
	static int  vdepths[] = { 8,16,15,32,24,4,2,1 };		// possible visual pixel size in pref. order
	static char pdepths[] = { 8,16,24,32, 0 };				// supported pixmap pixel size

	if (XLOG)						// list them all
	for ( j=0; j<(int)NELEM(visuals); j++ )
	for ( i=1; i<=32; i++ )  
	{	if ( XMatchVisualInfo ( display, DefaultScreen(display), i, visuals[j], &visiinf) )
		{	Log ( "available visual: %s with %i (%i) bits per pixel", visinam[j], visiinf.depth, i );
			if (visiinf.visual==DefaultVisualOfScreen(screen)) Log("(default visual of screen)");
		}
	}
	f = true;
	for ( i=0; i<(int)NELEM(vdepths)&&f;  i+=f )	  		// find best supported
	for ( j=0; j<(int)NELEM(visuals)&&f; j+=f )
	{
		f = ! ( XMatchVisualInfo ( display, DefaultScreen(display), vdepths[i], visuals[j], &visiinf ) 
		        && strchr(pdepths,(char)visiinf.depth) ); 
	}
	if (f) { Log("no supported visual"); done=true; goto x; }

	visual     = visiinf.visual;
	pixelsize  = visiinf.depth;
	visiclass  = visiinf.c_class;
	colordepth = vdepths[i];

	XLog( "choosen: %s with %i (%i) bpp", visinam[j], pixelsize, colordepth );

	pixelmask  = visiinf.red_mask|visiinf.green_mask|visiinf.blue_mask;
	if (visiclass==PseudoColor) pixelmask=~(0xFFFFFFFF<<pixelsize);		
	XTRAP(!pixelmask);

	if (XSAFE)						// verify pixmap depth IS supported
	{	XPixmapFormatValues *pixmaps;
		int numformats;
		pixmaps = XListPixmapFormats(display, &numformats);
		f = false;
		for ( i=0; i<numformats&&!f; i++ ) f = pixmaps[i].depth==pixelsize;
		if (!f) 
		{	Log("pixmap size for %s with %i bpp not supported (?!?!!)", visinam[j], pixelsize );
			done=true; goto x; 
		}
	}
	}

// initialize color tables and request colors from X
	CreateColors();
	

// create graphics context
	gc = XCreateGC ( display, root, 0, NULL );
	XCopyGC ( display, DefaultGCOfScreen(screen), ~0UL, gc );
	XSetGraphicsExposures ( display, gc, False );


// calculate initial geometry (size & position) of window
	{
	int sn = XScreenNumberOfScreen(screen);
	int sw = DisplayWidth(display,sn);
	int sh = DisplayHeight(display,sn);

	width  = WIN_STDWIDTH;	// default width
	height = WIN_STDHEIGHT;	// default height
	xpos   = (sw-width)/2;	// center window
	ypos   = (sh-height)/2;	// center window

	xim    = NULL;			// marker: no offscreen image allocated !
							// image must be created by calling SetScale()


// create window
	XSetWindowAttributes wattr;
	wattr.event_mask = KeyPressMask | KeyReleaseMask | ExposureMask | EnterWindowMask | LeaveWindowMask | StructureNotifyMask;
	wattr.colormap = cmap;
	Long wmask = CWEventMask | CWColormap;
	mainWin = XCreateWindow(display, root, xpos, ypos, width, height, 0, pixelsize, InputOutput, visual, wmask, &wattr);
	}


// notify window manager
	{
	XTextProperty	appName;
	XTextProperty	iconName;
	Pixmap		iconPix;
	XSizeHints	xsh;
	XWMHints   	xwmh;
	XClassHint 	xch;

	OMEM ( XStringListToTextProperty((char**)&applName, 1, &appName ) );
	OMEM ( XStringListToTextProperty((char**)&applName, 1, &iconName) );

	iconPix = XCreatePixmapFromBitmapData 
		( display, root, zxsp_bits, zxsp_width, zxsp_height, black, white, pixelsize );

	xsh.flags = PMinSize | PMaxSize | PResizeInc /*| PAspect*/ /*| PBaseSize */
		    | (1 ? USPosition : PPosition);

	xsh.min_width      = WIN_MINWIDTH;
	xsh.min_height     = WIN_MINHEIGHT;
	xsh.max_width      = WIN_MAXWIDTH;
	xsh.max_height     = WIN_MAXHEIGHT;

//	xsh.min_aspect.x   =
//	xsh.max_aspect.x   =
	xsh.width_inc      = WIN_GRID;
//	xsh.min_aspect.y   =
//	xsh.max_aspect.y   =
	xsh.height_inc     = WIN_GRID;

//	xsh.base_width     = 0;
//	xsh.base_height    = 0;

	xwmh.initial_state = NormalState;
	xwmh.input         = True;
	xwmh.icon_pixmap   = iconPix;
	xwmh.flags         = StateHint | IconPixmapHint | InputHint;

	xch.res_name       = "zxsp";
	xch.res_class      = "Zxsp";

	XSetWMProperties ( display, mainWin, &appName, &iconName, argv, argc, &xsh, &xwmh, &xch );
	XFree(appName.value);
	XFree(iconName.value);

	killme = XInternAtom ( display, "WM_DELETE_WINDOW", False );
	XSetWMProtocols ( display, mainWin, &killme, 1 );
	}

// open window
	XMapRaised ( display, mainWin );

x:	XLogExit();
}




void Win::ConfigureEvent ( XEvent& xev )
{
	XXLogLine ( ">Win::Event(ConfigureNotify)" );

	int w = xev.xconfigure.width;
	int h = xev.xconfigure.height;

	if (w!=width||h!=height)
	{
		XXLog ( "window resized" );	

		XTRAP(w<WIN_MINWIDTH);
		XTRAP(h<WIN_MINHEIGHT);
		XTRAP(w%WIN_GRID!=0);
		XTRAP(h%WIN_GRID!=0);

		DestroyImage();
		width=w; height=h; xim=0;
		// offscreen image must be recreated by calling SetScale() !
	}

	XXLogExit();
}


/* ----	poll events from X windows system ---------------------------------
*/
void zxsp::PollEvents( )
{
	static XEvent xev;

	XXXLogLine ( ">zxsp::PollEvents()" );

	while ( screen.PollEvents(xev) )
	{
		switch(xev.type)
		{
		case EnterNotify:
			if (xev.xcrossing.detail != NotifyInferior) { }
			break;
		case LeaveNotify:
			if (xev.xcrossing.detail != NotifyInferior) keyboard.Reset();
			break;
		case KeyPress:
		case KeyRelease:
			keyboard.KeyEvent((XKeyEvent *)&xev);
			break;
		}
	}

	XXXLogExit();
}


/* ----	poll events from X windows system ---------------------------------
*/
bool Scr::PollEvents( XEvent& xev)
{	bool f=false;

	XXXLogLine ( ">Scr::PollEvents()" );

	while ( !f && XEventsQueued ( window.Disp(), QueuedAfterFlush) )
	{
		XNextEvent ( window.Disp(), &xev );

		if (XXLOG)
		{
			static const char *eventName[] = 
			{
				"",					"",		
				"KeyPress",			"KeyRelease",
				"ButtonPress",		"ButtonRelease",	
				"MotionNotify",		"EnterNotify",
				"LeaveNotify",		"FocusIn",	
				"FocusOut",			"KeymapNotify",		
				"Expose",			"GraphicsExpose",
				"NoExpose",			"VisibilityNotify",		
				"CreateNotify",		"DestroyNotify",	
				"UnmapNotify",		"MapNotify",		
				"MapRequest",		"ReparentNotify",
				"ConfigureNotify",	"ConfigureRequest",	
				"GravityNotify",	"ResizeRequest",
				"CirculateNotify",	"CirculateRequest",	
				"PropertyNotify",	"SelectionClear",
				"SelectionRequest",	"SelectionNotify",	
				"ColormapNotify",	"ClientMessage",
				"MappingNotify"
			};
			assert((uint)xev.type < NELEM(eventName));
			Log ( "xev=%s", eventName[xev.type] );
		}

		switch(xev.type)
		{
		case Expose:
			window.FullUpdate();
			XXLog("requesting full screen update");
			break;
		case ConfigureNotify:
			window.ConfigureEvent(xev);
			break;
		case EnterNotify:
		case LeaveNotify:
		case KeyPress:
		case KeyRelease:
			f=true; break;	// => keyboard
		case ClientMessage:
			done |= xev.xclient.format==32 && (Atom)xev.xclient.data.l[0]==window.Killme();
			break;
		}
	}

	XXXLogExit();
	return f;
}


/* ----	colorize whole long from single pixel --------------------------
		except for 32bpp more than one pixel fit in 4 bytes of video memory.
		this function fills the whole long with the pixel's ink.
		odd man out: 24 bpp 
*/
static void SpreadPixel ( int pixelsize, Long& pixel )
{
	Long n = pixel;

	switch ( pixelsize )
	{
	case 24:n |= n&0x000000ff ? n<<24: n>>24;  break;
	case 1: n |= n&0x01010101 ? n<<1 : n>>1;
	case 2:	n |= n&0x03030303 ? n<<2 : n>>2;		
	case 4:	n |= n&0x0f0f0f0f ? n<<4 : n>>4;		
	case 8:	n |= n&0x00ff00ff ? n<<8 : n>>8;		
	case 16:n |= n&0x0000ffff ? n<<16: n>>16;  break;
	}

	pixel = n;
}


/* ---	calculate difference between colors -------------------------
*/
static long DeltaColor ( XColor& c1, XColor& c2 )
{	long r,g,b;

	r = ( (long)c1.red   - (long)c2.red   )/256;
	g = ( (long)c1.green - (long)c2.green )/256;
	b = ( (long)c1.blue  - (long)c2.blue  )/256;

	return r*r+g*g+b*b;
}


/* ----	calculate XColors for specci colors --------------------------
*/
static void CalculateSpecciColors ( XColor rgb[WIN_COLORS] )
{
	for (int i=0;i<WIN_COLORS;i++)
	{
		Short c = i>=8 ? 0xffff : 0xdddd;
		rgb[i].red   = i&2 ? c : 0;
		rgb[i].green = i&4 ? c : 0;
		rgb[i].blue  = i&1 ? c : 0;
		rgb[i].flags = DoRed|DoGreen|DoBlue;

	}
}


void Win::CreateColors ( )
{	int i,j;
	XColor rgb[WIN_COLORS];

	XLogLine ( ">CreateColors()" );
	XTRAP(cmap);
	XXTRAP(WIN_COLORS!=16);

	if ( visual == DefaultVisualOfScreen(screen) )
	{
	// default visual => try to use default colormap
		XXLog("try to use default colormap");
		cmap = DefaultColormapOfScreen(screen);

	// calculate XColors for all specci colors
		CalculateSpecciColors(rgb);

	// allocate colors
		for ( i=0; i<WIN_COLORS; i++ ) if ( XAllocColor(display,cmap,&rgb[i]) ) color[i]=rgb[i].pixel; else break;
		if (i==WIN_COLORS) goto x;	// got 'em all? => exit

	// we didn't get all colors
		XXLog ( "didn't get %i of 16 colors", WIN_COLORS-i );
		if (i)  XFreeColors(display, cmap, color, i-1, 0);
	//	int freecolors = i;

	// search best fit
		// get current colors
		XXXLog("get current colors");
		XColor colors[visiinf.colormap_size];
		for(i=0;i<visiinf.colormap_size;i++) colors[i].pixel=i;
		XQueryColors(display,cmap,colors,visiinf.colormap_size);

		// search best fit for each specci color
		XXXLog("searching best fits");
		#define ACCEPTABLE 3*20*20

		for (i=0;i<WIN_COLORS;i++)
		{
			long dif = 0x7fffffff;
			long idx = 0;

			for (j=visiinf.colormap_size-1; j>=0; j--)
			{
				long d = DeltaColor(rgb[i],colors[j]);
				if (d<dif) { dif=d; idx=j; }
			}
			XXLogLine("best match for %04X/%04X/%04X is %04X/%04X/%04X (diff=%i)", 
				rgb[i].red,rgb[i].green,rgb[i].blue,
				colors[idx].red,colors[idx].green,colors[idx].blue, dif );
			if (dif<ACCEPTABLE) rgb[i] = colors[idx];
		}

	// matches are acceptable: use 'em
		XXXLog("allocating approximated colors");
		for ( i = 0; i < WIN_COLORS; i++ ) if ( XAllocColor(display,cmap,&rgb[i]) ) color[i]=rgb[i].pixel; else break;
		if (i==WIN_COLORS) goto x;	// got 'em all => exit

		XLog("allocation of color %04X/%04X/%04X failed", rgb[i].red,rgb[i].green,rgb[i].blue);
		if (i)  XFreeColors(display, cmap, color, i-1, 0);
	}


	// we need a private colormap

	XXLog("using private colormap");
	cmap = XCreateColormap ( display, root, visual, AllocNone );

	// calculate XColors for all specci colors
	CalculateSpecciColors(rgb);

	// allocate them
	for ( i=0; i<WIN_COLORS; i++ ) if ( XAllocColor(display, cmap, &rgb[i]) ) color[i]=rgb[i].pixel;
	XTRAP(i!=WIN_COLORS);

x:	// spread pixels to cover whole longs
	for (i=0;i<WIN_COLORS;i++) SpreadPixel(Depth(),color[i]);

	XLogExit();
}


void Win::DestroyColors ( )
{
	XXLogLine ( ">Win::DestroyColors()" );

	if (cmap)
	{
		XFreeColors ( display, cmap, color, WIN_COLORS, 0 );
		if ( cmap != DefaultColormapOfScreen(screen) ) XFreeColormap(display, cmap);
	}

	XXLogExit();
}





/* ----	handle error when attaching MIT SHM -----------------
		(( Global! ))
*/
#ifdef HAVE_MITSHM
	static bool 		attachError;
	static XShmSegmentInfo	xshminfo;
	static int 		(*oldErrorHandler)(Display*, XErrorEvent*);

	static int CatchAttachError(Display *disp, XErrorEvent *ev)
	{
		XLogLine ( "CatchAttachError(disp,ev)" );

		if (ev->error_code == BadAccess)
		{
			attachError = 1;
			return 0;
		}
		return (oldErrorHandler(disp, ev));
	}
	#if defined(nec_ews) && !defined(__STDC__)
		void *shmat(int, void*, int);
	#endif
#endif


/* ----	create offscreen image -------------------------------
	global in:	display, visual, depth
	global out:	xim
*/
void Win::CreateImage ( int w, int h )
{
	XXLogLine ( ">Win::CreateImage(%i,%i)", w, h );
	XTRAP(xim!=NULL);

#ifdef HAVE_MITSHM
	if (mitshm)
	{
		int x, y, z, f;
		if ( XQueryExtension ( display, "MIT-SHM", &x, &y, &z ) )
		{
			XShmQueryVersion ( display, &x, &y, &f );
			mitshm = f;
		}
		else	mitshm = false;
	}

	if (mitshm)
	{
		xim = XShmCreateImage(display, visual, pixelsize, ZPixmap, NULL, &xshminfo, w, h );
		if (!xim) goto error;
		xshminfo.shmid = shmget(IPC_PRIVATE, xim->bytes_per_line * xim->height, IPC_CREAT | 0777);
		if (xshminfo.shmid == -1) goto error;
		xshminfo.shmaddr = xim->data = (ptr)shmat(xshminfo.shmid, 0, 0);
		if (!xshminfo.shmaddr) goto error;
		xshminfo.readOnly = False;
		attachError = 0;
		oldErrorHandler = XSetErrorHandler(CatchAttachError);
		if (!XShmAttach(display, &xshminfo) || attachError)
		{
			XSetErrorHandler(oldErrorHandler);
			goto error;
		}
		XSetErrorHandler(oldErrorHandler);
		shmctl(xshminfo.shmid, IPC_RMID, 0);
		goto x;

error:
		Log ( "couldn't create MIT-SHM image");

		if (xim)
		{
			if (!attachError)	XShmDetach(display, &xshminfo);
			XDestroyImage(xim); xim=0;
			if (xshminfo.shmaddr)  	shmdt(xshminfo.shmaddr);
			if (xshminfo.shmid!=-1)	shmctl(xshminfo.shmid, IPC_RMID, 0);
		}
		mitshm = false;
	}
#endif

	OMEM ( xim = XCreateImage(display, visual, pixelsize, ZPixmap, 0, NULL, w, h, 8, 0) );
	OMEM ( xim->data = (char*)malloc(xim->bytes_per_line * xim->height) );

x:	fullupdate=true;	// invalidate whole screen

	XXLogExit();
}


void Win::DestroyImage ( )
{
	XXLogLine ( ">Win::DestroyImage()" );

	if (xim)
	{
		xim->data = data;
		xim->bytes_per_line = yoffs;

#ifdef HAVE_MITSHM
		if (mitshm)
		{
			XShmDetach(display, &xshminfo);
			XDestroyImage(xim);
			shmdt(xshminfo.shmaddr);
		}
		else
#endif
		{
			XDestroyImage(xim);
		}
	}

	xim=0;

	XXLogExit();
}


/* ----	select new magnification scale for specci screen -------------
	if still valid, the current offscreen image is freed 
	and a new offscreen image for the current window width and height 
	and the requested magnification scale is allocated.

note:	application's view: the offscreen image is scaled up horizontally according 
	to scale but it is *not* scaled vertically!
	viewed from the window, the offscreen image is not scaled horizontally
	but scaled down vertically.

	since with magnification scale N every N scan lines are identical, there is 
	no need to store the duplicate lines if we handle this in Win::Update()

	so the application has to scale the image horizontally when drawing into
	the offscreen image, but not vertically.

	Win::Update() has to scale the image data vertically when posting the updates
	to the X server, but not horizontally.

	Coordinate parameters are treated as normalized to scale=1 in all functions 
	taking such parameters e.g. Win::Update() and Win::Data(), 
	and they are returning values normalized to scale=1 as in Width() and Height(), 
	if not returning 'real' sizes as in RealWidth() and RealSize().
*/
void Win::SetScale ( int newscale )
{ 
	XLogLine(">Win::SetScale(%i)",newscale);

	if (xim) DestroyImage(); 

	scale = newscale;
	int h  =  1+(height-1)/scale;
	int w  = (1+(width -1)/scale) *scale;

	CreateImage(w,h); 

	data = xim->data;
	yoffs = xim->bytes_per_line;
	xim->bytes_per_line = 0;
	fullupdate = true;

	XLogExit();
}


/* ----	update horizontal line within image ---------------------------------------
	parameters are normalized to scale=1
	the requested line is posted to the X servers output queue. 
	it is probably not drawn imediately. 
	to force the update, call Win::Sync() after having called Win::Update() 
	for all changed pixels.

note:	the offscreen image is horizontally scaled according to scale
	but it is *not* scaled vertically!
	we fake a horizontally and vertically scaled image by setting xim->bytes_per_line to 0. 
	so, when XPutImage() tries to step to the next scanline, it gets the old scanline again.
	this only works with single horizonal lines, not with blocks of rows. 
	that's why Update() only takes single horizonal lines.
*/
void Win::Update ( int x, int y, int width )
{
	if (fullupdate) return;		// already full screen update requested

	XXXLogLine(">Win::Update(%i,%i,%i)", x,y,width);

	XTRAP(xim==NULL);
	XTRAP(xim->bytes_per_line!=0);

	x*=scale;width*=scale;

	XXTRAP(x<0);
	XXTRAP(y<0);
	XXTRAP(width<=0);
	XXTRAP(x+width>xim->width);
	XXTRAP(y>xim->height);

	xim->data = data+y*yoffs;

#ifdef HAVE_MITSHM
	if (mitshm)
		XShmPutImage
		( display, mainWin, gc, xim, x, 0, x, y*scale, width, scale, false );
	else
#endif
		XPutImage
		( display, mainWin, gc, xim, x, 0, x, y*scale, width, scale );

	XXXLogExit();
}


/* ----	wait for updates to finish --------------------------
*/
void Win::Sync ( )
{
	XXLogLine(">Win::Sync()");

	if (fullupdate)
	{
		fullupdate = false;
		for (int y=0;y<Height();y++) Update(0,y,Width());
	}

	// flushing does not work well on my X server
	// the X server seems to buffer the retrieved commands internally
	// the window seems to be *not* up to date after XSync() !
	// the window is *not* as smoothly animated as it shoud be !
	//								kio 17.jan.99

	XSync ( display, False );	// flush output queue and wait for completion
	//XFlush( display );		// flush output queue

	XXLogExit();
}





