
/*	Z80 Emulator

***	Copyright:	1995-1999 Gnter Woigk
  				mailto:kio@myokay.net

 	This program is distributed in the hope that it will be useful,
 	but WITHOUT ANY WARRANTY; without even the implied warranty of
 	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  

 	Permission to use, copy, modify, distribute, and sell this software and
 	its documentation for any purpose is hereby granted without fee, provided
 	that the above copyright notice appear in all copies and that both that
 	copyright notice and this permission notice appear in supporting
 	documentation, and that the name of the copyright holder not be used
 	in advertising or publicity pertaining to distribution of the software
 	without specific, written prior permission.  The copyright holder makes no
 	representations about the suitability of this software for any purpose.
 	It is provided "as is" without express or implied warranty.

 	THE COPYRIGHT HOLDER DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS SOFTWARE,
 	INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS, IN NO
 	EVENT SHALL THE COPYRIGHT HOLDER BE LIABLE FOR ANY SPECIAL, INDIRECT OR
 	CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE,
 	DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER
 	TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR
 	PERFORMANCE OF THIS SOFTWARE.
			
	based on fMSX; Copyright (C) Marat Fayzullin 1994,1995

	16.Jan.95 	Fixed wrong calculation of parity flag (68k version)		KIO !
	28.Mar.95 	Started work on this file					KIO !
	01.Apr.95	Finished, yee it works!						KIO !
	06.Apr.95	Removed "final bug": im2 interrupt handling			KIO !
	12.Jun.95	revised EI_WITH_DELAY handling					KIO !
	12.Jun.95	all Info-Calls and profiling added				KIO !
	12.Jun.95	exact r register emulation completed				KIO !
	13.Jun.95	All opcodes covered, no unsupported opcodes!			KIO !
	01.Nov.95 	Paged memory supported						KIO !
	26.Nov.95	Engine now completely hardware independent!			KIO !
	28.Mar.96	Core2Core() works even if src and dest block overlap at both ends	KIO !
	28.Mar.96	removed option to switch off latency of 'EI' instruction		KIO !
	02.Mai.96	load_cc after IN and OUT added to make extra wait states possible	KIO !
	24.Dec.98	started port to linux, rework for c++				kio !
*/

#include "z80.h"
#include "macros.h"
#include "opcodes.h"


#ifndef	LOG
#define	LOG	1
#endif

#ifndef	SAFE
#define	SAFE 	2
#endif


// ==================================================================

// ----	virtual destructor of class micro -------------------------
micro::~micro()
{
}


// ==================================================================
//	DATA
// ==================================================================

static Char	zlog_flags[256];		// convert:   A -> Z80-flags with V=parity, S, Z, and C=0
static Char	noreadpage[pagesize]; 	// dummy page for unmapped memory access
static Char	nowritepage[pagesize];	// ditto


static void InitZ80Tables ( )
{
// read from non-mapped ram: 
	const Char nobyte = 0xFF;
	MemSet ( nobyte, noreadpage, pagesize );

// z80 flag tables:		%SZ-H-VNC	(S, Z and P only)
	for (int i=0; i<256; i++)
	{	
		zlog_flags[i] =	(i&0x80)  +  ((i==0)<<6)  + 
				(  ( (~i+(i>>1)+(i>>2)+(i>>3)+(i>>4)+(i>>5)+(i>>6)+(i>>7))&1 ) << 2  );
	};

}


// ==================================================================================


/* ---------------------------------------------------------
	creator:
	init static tables
	init memory page pointers (no ram)
	init z80 registers (reset)
--------------------------------------------------------- */
z80::z80 ( micro& board )
:	myboard (board)
{
	XLogLine ( ">z80::z80(myboard)" );

	InitZ80Tables();
	for (int i=0;i<corepages;i++)  AttachRam ( i,(unsigned char*)nil );
	Reset();

	XLogLine("<");
}


/* ---------------------------------------------------
	Attach real memory to z80 address space
	for reading if rwm&1, writing if rwm&2 and opcode fetch M1 cycle if rwm&4

	page[i] pointers do NOT point to the start of a page.
	they point to the virtual location of address $0000 instead!
	so you can access a byte:  ptr = page[addr>>pagebits] + addr
--------------------------------------------------- */

void z80::AttachPage ( int idx, Ptr page, int rwm )
{
	XXLogLine ( ">z80::AttachPage(%i,%08X,%i)", idx, page, rwm );

	XTRAP ( idx<0 || idx>=corepages );

	if (page)
	{
		page = page - idx*pagesize;
		if (rwm&1) rpage[idx] = page;
		if (rwm&2) wpage[idx] = page;
		if (rwm&4) mpage[idx] = page;
	}
	else
	{
		if (rwm&1) rpage[idx] = noreadpage -idx*pagesize;
		if (rwm&2) wpage[idx] = nowritepage -idx*pagesize;
		if (rwm&4) mpage[idx] = noreadpage -idx*pagesize;
	}

	XXLogLine("<");
}


/* ---------------------------------------------------
	reset z80 cpu
--------------------------------------------------- */
void z80::Reset ( )
{
	XLog("z80::Reset()");

	BC=DE=HL=IX=IY=PC=SP=BC2=DE2=HL2=0;	// clear registers
	RA=RA2=RF=RF2=RR=RI=0;
	IFF1=IFF2 = disabled;			// disable interrupts
	IM = 1;					// interrupt mode := 1
	NMI = IRPT = 0;				// clear irpt, nmi
	IRPTCMD = 0xff;				// byte read from bus in im0 or im2 mode
	cycles = 0;				// T cycle down counter
}


/* ----------------------------------------------------------------
	Core to core copying
	These functions implement the Z80 LDIR and LDDR instruction
---------------------------------------------------------------- */

void z80::Do_LDIR ( Short src, Short dest, Short len )
{	Short cnt; Char *q,*z;

	XXXLog("z80::Do_LDIR()");

	do
	{	q = RdPtr(src); z = WrPtr(dest);
		cnt = pagesize-Max(src&pagemask,dest&pagemask);
		if (cnt>len)  cnt = len?len:cnt;
		src+=cnt; dest+=cnt; len-=cnt;
		while (cnt--) *z++=*q++;
	} 
	while (len);
}


void z80::Do_LDDR ( Short src, Short dest, Short len )
{	Short cnt; Char *q,*z;

	XXXLog("z80::Do_LDDR()");

	do
	{	q = RdPtr(src); z = WrPtr(dest);
		cnt = 1+Min((src-1)&pagemask,(dest-1)&pagemask);
		if (cnt>len)  cnt = len?len:cnt;
		src-=cnt; dest-=cnt; len-=cnt;
		while (cnt--) *z--=*q--;
	} 
	while (len);
}



static char* flagstr(Char f)
{	static char s[]="SZ-H-VNC";

	s[0] = f&0x80 ? 'M' : 'P';
	s[1] = f&0x40 ? 'Z' : '-';
	s[3] = f&0x10 ? 'H' : '-';
	s[5] = f&0x04 ? 'V' : '-';
	s[6] = f&0x02 ? 'N' : '-';
	s[7] = f&0x01 ? 'C' : '-';

	return s;
}


// ===========================================================================
// ====	The Z80 Engine =======================================================
// ===========================================================================


// ----	jumping -----------------------------------------------------------
#define	loop		goto nxtcmnd			// loop to next instruction
#define	exit(N)		{ c=N; goto exit_z80; }		// exit z80


// ----	load and store local variables from/to global registers ------------
#define	store_cc	cycles = cc
#define	load_cc		cc = cycles

#define	LOAD_REGISTERS					\
	load_r;			/* refresh counter R	*/	\
	load_cc;		/* T cycle counter	*/	\
	pc = PC;		/* program counter PC	*/	\
	sp = SP;		/* stack pointer SP	*/	\
	ra = RA;		/* register A		*/	\
	rf = RF;		/* register F		*/

#define	SAVE_REGISTERS					\
	store_r;		/* refresh counter R	*/	\
	store_cc;		/* T cycle counter	*/	\
	PC = pc;		/* program counter PC	*/	\
	SP = sp;		/* stack pointer SP	*/	\
	RA = ra;		/* register A		*/	\
	RF = rf;		/* register F		*/


/* ====	The Z80 ENGINE ====================================================
*/
OSErr z80::Run ( long moreT )
{	

register long	cc;					// T cycle counter

register Short	pc;					// z80 program counter
register Short	sp;					// z80 stack pointer
register Char	ra;					// z80 a register
register Char	rf;					// z80 flags

#define		time(N)	cc-=N-4				// T cycles for instruction
#define		more(N)	cc-=N				// adjust T cycles after use of time()

register pair	*izp;					// pointer to IX or IY after 0xDD or 0xFD
#define		iz		(izp->rr)		// IX or IY
#define		rzh		(izp->r.hi)		// XH or YH
#define		rzl		(izp->r.lo)		// XL or YL

register Char	c;					// general purpose byte register
	
register Short	w;					// general purpose word register
#define		wl		(Char)w			// access low byte of w
#define		wh		(w>>8)			// access high byte of w

register Short	wm;					// scratch for macro internal use
#define		wml		(Char)wm		// access low byte of wm
#define		wmh		(wm>>8)			// access high byte of wm

long		ccx = 0;					// limit for cc -> 1 instr. latency after ei handling

#if EXACT_R
register Char	r;					// r register bit 0...6
#endif

	XXLogLine ( ">z80::Run(%i)", moreT );

// load local variables from global registers:
	LOAD_REGISTERS;
	cc += moreT;


// check for interrupts:
// interrupts are only tested once at start of z80::Run()
// and only tested again after EI processing!
check_irpt:
	if (NMI)		// NMI:  0x0066; 11 T cycles
	{
		XXLog ("-NMI-");

		DO_INFO_NMI;
		NMI = false;
		IFF2=IFF1;
		IFF1=disabled;
		PUSHD(pc);
		pc = 0x0066;
		more(11);
		loop;
	}

	if (IRPT && IFF1==enabled)		// interrupts enabled?
	{
		XXLog ("-IRPT-");

		DO_INFO_IRPT;
		IRPT--;				// clear interrupt request
		IFF1=IFF2=disabled;		// disable iff1 & iff2
		PUSHD(pc);			// push actual pc

		switch (IM)
		{
		case 0:	// mode 0: read instr; 1 T cycle + normal instr timing
			if (IRPTCMD==RST38)
			{
				pc = 0x0038;
				more(12);	// RST==11T + 1T
				loop;
			};
			exit ( irpt_error );	// not supported instruction read in irpt ackn cycle

		case 1:	// mode 1:  rst 56; 13 T cycles
			pc = 0x0038;
			more(13);
			loop;

		case 2:	// mode2:  jmp via table; 19 T cycles
			w = IRPTVEK;
			pc = PEEK(w+1)*256 + PEEK(w);
			more(19);
			loop;

		default:
			exit ( irpt_error );
		};
	};


// ----	THE MAIN INSTRUCTION DISPATCHER --------------------------------

nxtcmnd:

	if (cc>ccx)			
	{	
		if (XXXLOG) 
		{
			if (pc<8) { done=true; cycles=0; exit(ok); }
		//	LogLine ( "\t   A    BC    DE    HL    IX    IY    SP SZ-H-VNC   R   I IM EI" );
		//	LogLine ( "\t%4i%6u%6u%6u%6u%6u%6u %s%4i%4i%3i%3i", ra,BC,DE,HL,IX,IY,sp, flagstr(rf), r, RI, IM, 0+(IFF1==enabled) );
		//	PC=pc; LogLine ( "%04X: %s", pc, Disassemble(&PC) ); 
			LogLine ( " A:%02X BC:%04X DE:%04X HL:%04X IX:%04X IY:%04X SP:%04X %s R:%02X I:%02X %i:%s", ra,BC,DE,HL,IX,IY,sp, flagstr(rf), r&0x7f+RR&0x80, RI, IM, IFF1==enabled?"EI":"DI" );
		 	PC=pc; LogLine ( "$%04X:\t%s", pc, Disassemble(&PC) ); 
		}
		#include "codes.c"
	}
	else
	{
		if (ccx)	ccx=0;		// for correct ei handling
		else	cc=myboard.Cycles(cc);	// please reload counter ...
		if (cc>0) goto check_irpt;	// counter reloaded ?
		exit ( tcycles_expired );	// else exit
	};

exit_z80:
	SAVE_REGISTERS;

	XXLogLine("<");

	return c;
}






