/*	Z80 Emulator: header file

***	Copyright:	1995-1999 Gnter Woigk
  				mailto:kio@myokay.net

 	This program is distributed in the hope that it will be useful,
 	but WITHOUT ANY WARRANTY; without even the implied warranty of
 	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  

 	Permission to use, copy, modify, distribute, and sell this software and
 	its documentation for any purpose is hereby granted without fee, provided
 	that the above copyright notice appear in all copies and that both that
 	copyright notice and this permission notice appear in supporting
 	documentation, and that the name of the copyright holder not be used
 	in advertising or publicity pertaining to distribution of the software
 	without specific, written prior permission.  The copyright holder makes no
 	representations about the suitability of this software for any purpose.
 	It is provided "as is" without express or implied warranty.

 	THE COPYRIGHT HOLDER DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS SOFTWARE,
 	INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS, IN NO
 	EVENT SHALL THE COPYRIGHT HOLDER BE LIABLE FOR ANY SPECIAL, INDIRECT OR
 	CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE,
 	DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER
 	TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR
 	PERFORMANCE OF THIS SOFTWARE.
  
	class z80
*/

#ifndef	z80_h
#define	z80_h z80_h

#include "kio.h"
#include "options.h"


/* -----------------------------------------------------------
	virtual class for micro computer
	you shoud derive a class for your real computer implementation
----------------------------------------------------------- */

class micro
{
public:
virtual	~micro		( );

// call back functions:
virtual void	Output		( long cc, Short addr, Char n )	= 0;	// Output byte to port
virtual Char	Input 		( long cc, Short addr )		= 0;	// Input byte from port
virtual long	Cycles		( long cc )			= 0;	// Count down for M1 cycles reached 0; request for reload
// the following functions are optional ((see "z80.options"))
virtual void	Info_Misc	( long cc, Short ip )	= 0;		// info call for interesting instruction
virtual void	Info_Illegal	( long cc, Short ip )	= 0;		// info call for illegal instruction
virtual void	Info_Weird	( long cc, Short ip )	= 0;		// info call for weird illegal instruction
virtual void	Info_Irpt		( long cc, Short ip )	= 0;		// info call if INFO_IRPT==on
virtual void	Info_NMI	( long cc, Short ip )	= 0;		// info call if INFO_NMI==on
virtual void	Info_1st_Loc	( long cc, Short ip )	= 0;		// info call if PC_PROFILE==on
virtual void	Info_1st_Instr	( long cc, Short ip )	= 0;		// info call if CMD_PROFILE==on
};


/* -----------------------------------------------------------
	z80 cpu class
------------------------------------------------------------*/

enum	{ legal, illegal, weird };		// legal state of instructions

enum	{ disabled=0x00, enabled=0xFF };	// interrupt flag constants

enum				// return values of z80::Run()
{	tcycles_expired = 0,	// noerror
	halt_instr,		// halt executed
	trap_instr,		// trap instruction (( rst0 ))
	irpt_error,		// unsupported instruction read in irpt mode 0
	nimp_instr,		// not implemented instruction (( this cannot happen ))
	internal_error		// other internal error (( this cannot happen ))
};


// ---- z80.options derived data ----------------------------------------------

const int  corepages       	= 1<<(16-pagebits);     	// pages inside Z80 address space
const int  pagesize        	= 1<<pagebits;          	// e.g.: 0x4000
const int  pagemask        	= pagesize-1;     	// e.g.: 0x3fff


// ----	Z80 register pair ---------------------------------------------------

union pair
{
    Short			rr;
    struct {Char lo,hi;}	r;		// big endian (intel)
};


// ---- class for z80 cpu ---------------------------------------------------

class z80
{

    micro	  	&myboard;		// the micro computer this z80 cpu is soldered in :-)

    long		cycles;			// processor T cycles (count down for interrupt)

// the z80 registers:
    pair		af,bc,de,hl;		// 1st register set
    pair		af2,bc2,de2,hl2;	// 2nd register set
    pair		ix,iy,pc,sp;		// special registers
    pair		iic;			// register i + hibyte read from bus
    Char		iff1,iff2;		// interrupt enable flip flops
    Char		nmi,irpt;		// interrupt flags
    Char		im;			// interrupt mode (0 ... 2)
    Char		r;			// refresh counter (7 bit)

// the z80 core:
    Char		*rpage[corepages];	// mapped pages for reading
    Char		*wpage[corepages];	// mapped pages for writing
    Char		*mpage[corepages];	// mapped pages for instruction reading (m1 cycle)

// internal utilities:
    void		Do_LDIR	( Short src, Short dest, Short len );
    void		Do_LDDR	( Short src, Short dest, Short len );

public:
#if CMD_PROFILE			// instruction profiling:
    bool		count_instr;
    long		cnt_xx[256], cnt_cb[256], cnt_ed[256], cnt_xy[256], cnt_xycb[256];
#endif

#if PC_PROFILE				// M1 address profiling:
    bool		count_pc;
    long		cnt_pc[0x10000];
#endif


public:
// standard member functions:	
		z80		( micro& myboard );	// default creator
		~z80		( )	 { };		// destructor
		z80		( const z80& );		// prohibit
    z80&		operator=	( const z80& );		// prohibit

// attach memory:
    void		AttachPage	( int idx, Ptr page, int i );	// i: 1+2+4 = read+write+m1
    void		AttachRam	( int idx, Ptr page )	{ AttachPage(idx,page,1+2+4); };	// attach Ram to z80 
    void		AttachRom	( int idx, Ptr page )	{ AttachPage(idx,page,1+0+4); };	// attach Rom to z80
    void		AttachWom	( int idx, Ptr page )	{ AttachPage(idx,page,0+2+0); };	// attach write-through Ram
    void		AttachM1	( int idx, Ptr page )	{ AttachPage(idx,page,0+0+4); };	// attach invisible patch Rom

// run the cpu:
    OSErr	Run		( long moreT=0 ); 				// run the cpu
    void		Reset		( );						// reset the cpu
    void		NMI		( )			{ nmi=1; };		// trigger nmi
    void		Irpt		( Char busbyte=255 )	{ irpt=1;iic.r.lo=busbyte; };	// trigger irpt

// access registers:
// note: within the callback functions some registers are in register variables: at least AF, PC, SP !!!
    long&	Cycles		( )			{ return cycles; }
    Char&	IFF1		( )			{ return iff1; }
    Char&	IFF2		( )			{ return iff2; }
    Short&	PC		( )			{ return pc.rr; }
    Char&	PCH		( )			{ return pc.r.hi; }
    Char&	PCL		( )			{ return pc.r.lo; }
    Short&	SP		( )			{ return sp.rr; }
    Char&	SPH		( )			{ return sp.r.hi; }
    Char&	SPL		( )			{ return sp.r.lo; }
    Short&	BC		( )			{ return bc.rr; }
    Char&	RB		( )			{ return bc.r.hi; }
    Char&	RC		( )			{ return bc.r.lo; }
    Short&	DE		( )			{ return de.rr; }
    Char&	RD		( )			{ return de.r.hi; }
    Char&	RE		( )			{ return de.r.lo; }
    Short&	HL		( )			{ return hl.rr; }
    Char&	RH		( )			{ return hl.r.hi; }
    Char&	RL		( )			{ return hl.r.lo; }
    Char&	RA		( )			{ return af.r.hi; }
    Char&	RF		( )			{ return af.r.lo; }
    Char&	RA2		( )			{ return af2.r.hi; }
    Char&	RF2		( )			{ return af2.r.lo; }
    Short&	BC2		( )			{ return bc2.rr; }
    Short&	DE2		( )			{ return de2.rr; }
    Short&	HL2		( )			{ return hl2.rr; }
    Short&	IX		( )			{ return ix.rr; }
    Short&	IY		( )			{ return iy.rr; }
    Char&	IM		( )			{ return im; }
    Char&	RR		( )			{ return r; }
    Char&	RI		( )			{ return iic.r.hi; }
//  void	SetRegisters	( Short w[12],  /* af,bc,de,hl,af2,bc2,de2,hl2,ix,iy,pc,sp */
//				  Char  b[5],   /* i,r,im,iff1,iff2 */
//				  long  cc=0 ); /* cycles */

// access memory:  
    Char*	RdPtr 	( Short addr )			{ return  rpage[addr>>pagebits] + addr; }
    Char*	WrPtr 	( Short addr )			{ return  wpage[addr>>pagebits] + addr; }
    Char*	M1Ptr 	( Short addr )			{ return  mpage[addr>>pagebits] + addr; }
    Char	Peek  	( Short addr )			{ return  rpage[addr>>pagebits][addr]; }
    void	Poke  	( Short addr, Char byte )	{ wpage[addr>>pagebits][addr] = byte; }


// ----	disassembler stuff ----

// return mnenonic with symbolic arguments for instructions
	char*	OpMnemo	( Short ip );

// ----	calculate length of opcode
	int	OpLength	( Short ip );

// ----	querry legality of instruction
	int	OpLegal	( Short ip );

// ----	disassemble core position and advance pointer
	char*	_xword		( Char n, Short *ip );
	char*	Disassemble	( Short *ip);
};


extern	char* 	Z80OpMnemo	( Char op1, Char op2=0, Char op4=0 );
extern	int	Z80OpLength	( Char op1, Char op2=0 );
extern	int	Z80OpLegal	( Char op1, Char op2=0, Char op4=0 );

#endif















