/*
***	Copyright:	1998-1999 Gnter Woigk
  				mailto:kio@myokay.net

 	This program is distributed in the hope that it will be useful,
 	but WITHOUT ANY WARRANTY; without even the implied warranty of
 	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  

 	Permission to use, copy, modify, distribute, and sell this software and
 	its documentation for any purpose is hereby granted without fee, provided
 	that the above copyright notice appear in all copies and that both that
 	copyright notice and this permission notice appear in supporting
 	documentation, and that the name of the copyright holder not be used
 	in advertising or publicity pertaining to distribution of the software
 	without specific, written prior permission.  The copyright holder makes no
 	representations about the suitability of this software for any purpose.
 	It is provided "as is" without express or implied warranty.

 	THE COPYRIGHT HOLDER DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS SOFTWARE,
 	INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS, IN NO
 	EVENT SHALL THE COPYRIGHT HOLDER BE LIABLE FOR ANY SPECIAL, INDIRECT OR
 	CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE,
 	DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER
 	TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR
 	PERFORMANCE OF THIS SOFTWARE.
*/


#include	<stdio.h>
#include	<assert.h>
#include	<unistd.h>
#include	<malloc.h>
#include	<sys/mman.h>
#include	<stdlib.h>

#include	"options.h"
#include	"kio.h"
#include	"zxsp.h"
#include	"ay.h"
#include	"screen.h"
#include	"util.h"
#include	"main.h"

#ifndef LOG
#define	LOG	1	// 0=nix ... 2=debug
#endif

#ifndef	SAFE
#define	SAFE	2	// 0=nix ... 2=debug
#endif


/* ===========================================================
			STATIC DATA
=========================================================== */


static long 	n2n[24*8];	// convert: display line number <-> pixmap line index


/* ============================================================
			MEMBER FUNCTIONS
============================================================ */


/* ----	Send byte to ZX 128's paging control port ---------------------------
	Bit 0-2:Ram for top 16k:	0 is the usual bank;   7 contains some scratchpads used by the 128k editor.
	Bit 3:	Screen select:		0 => Use the normal screen in bank 5;	1 => Use the shadow screen in bank 7.
	Bit 4:	ROM for lower 16k:	0 => 128k editor;			1 => 48k BASIC
	Bit 5:	Lock. If set, disables all further paging.
*/
void zxsp::PageCtrl ( Char n )	
{	int i,j;

	if ( type!=zx128 ) return;		// not a ZX 128
	if (pagectrl&0x20) return;		// further paging disabled

	XXXLogLine ( ">PageCtrl($%02X)",n );

	if ( (n^pagectrl)&0x08 )		// video Ram select: page 5 or 7
	{
		i = 5 + ((n&0x8)>>2);
		XXXLog( " -> new video page #%i", i );
		videopage = ram[i];	
	}

	if ( (n^pagectrl)&0x07 )		// upper ram bank changes
	{
		XXXLog( " -> new upper ram bank #%i", n&0x07 );
		cpu.AttachRam ( 3, ram[n&0x07] );
	}
	
	if ( (n^pagectrl)&0x10 )		// rom select changes
	{
		rombits = (rombits&~2) + ((n>>3)&2);
		for ( j=0,i=rombits>>1; i; i>>=1 ) j++;
		XXXLog(" -> select rom #%i", j );
		cpu.AttachRom ( 0, rom[j] );
	}
		
	pagectrl = n;

	XXXLogExit();
}


/* ----	Attach a rom to the hardware -------------------------------
	the rom is not automatically paged in
*/
void zxsp::AttachRom ( int n, char* filename )
{
	XLogLine ( ">zxsp::AttachRom(%i,\"%s\")", n, filename );

	TRAP( n>=maxrompages );

	size_t size=pagesize;
	FILE* file = Fopen ( filename, "rb" );	// readonly, binary
	TRAP(!file);

	Ptr bu = (Ptr)mmap ( NULL, size, PROT_READ, MAP_SHARED, fileno(file), (off_t)0 );
	TRAP((long)bu<=0);

	fclose(file);

	MemCopy ( bu, rom[n], size );
	munmap((char*)bu,size);

	XLogExit();
}


/* ----	Select hardware ----------------------------------------
*/
void zxsp::SetType ( int typ )
{
	XLogLine ( ">zxsp::SetType(%i)", typ );

	type	= typ;

	rombits = 0x01;			// rom 0 enabled, roms 1, 2, 3 disabled
	cpu.AttachRom ( 0, rom[0] );
	
	framerate = 50;

	soundchip.EnableBeeper(true);

	switch ( type )
	{
	case zx16:
	case zx48:
		cpuclock = 3500000;
		linecycles = 224;
		topborderlines = 64;
		for ( int i=1; i<4;i++ )  cpu.AttachRam  ( i, ram[i] );
		videopage = ram[1];
		AttachRom(0,"48.rom");
		soundchip.EnableAY(false);
		break;
	
	case zx128:
		cpuclock = 3546900;
		linecycles = 228;
		topborderlines = 63;
		pagectrl = 0;
		cpu.AttachRam ( 1, ram[5] );
		cpu.AttachRam ( 2, ram[2] );
		cpu.AttachRam ( 3, ram[0] );
		videopage = ram[5];
		AttachRom(0,"128-0.rom");
		AttachRom(1,"128-1.rom");
		soundchip.EnableAY(true);
		break;
	default:
		IERR();
	}

// derived values:	
	framecycles = cpuclock/framerate;

// initialize the hardware
	Reset();

	XLogExit();
}


/* ----	press reset button ----------------------------------
	do anything i expect the hardware does
*/
void zxsp::Reset ( )
{
	XLogLine(">zxsp::Reset()");

	nextframe = frameCounter;
	endcycles = 0;
	scanline  = 0;

// reset page control register (if zx128)
	pagectrl &= ~0x20;
	PageCtrl(0);

// reset the cpu
	cpu.Reset();

// reset soundchip and beeper
	soundchip.Reset(cpuclock,cpuclock/2);

	XLogExit();
}


/* ----	create a new specci instance -----------------------
*/
zxsp::zxsp ( int& argc, char* argv[] ) 
:	cpu ( *this ),
	screen ( argc,argv ),
	soundchip(2000000,1000000)
{
	XLogLine ( ">zxsp::zxsp()" );

// table: physical screen row <-> specci screen row ordering
	for (int i=0;i<24*8;i++)  n2n[i] = ((i&0x07)<<3) + ((i&0x38)>>3) + (i&0xC0);

// roms:
	for (int i=0;i<maxrompages;i++)
	{
		assert ( rom[i] = new Char[pagesize] );	
	}

// rams:
	for (int i=0;i<maxrampages;i++)
	{
		assert ( ram[i] = new Char[pagesize] );

		Long *p = (Long*)(ram[i]);
		for (int j=0;j<pagesize/4;j++) *p++ = random();
	}
	
// screen and border
	border = 0;		// black
	videopage = ram[1];
	type = -1;
	rombits = 0x01;
	speaker = 0;
	cpuclock=2000000;	

	Reset();

	XLogExit();
}


zxsp::~zxsp()
{
	XLogLine("zxsp::~zxsp()");
	for (int i=0;i<maxrompages;i++)	delete[] rom[i];	
	for (int i=0;i<maxrampages;i++)	delete[] ram[i];
}


void zxsp::Irpt ( Char busbyte )
{
	cpu.Irpt(busbyte);
}


void zxsp::NMI ( )
{
	cpu.NMI();
}


long zxsp::Cycles( long cc )
{
	XXXLogLine ( ">zxsp::Cycles(%i)", cc );

a:	int line = scanline++;

	XXXLog ( "visible scanline #%i (max:%i)", line, screen_lines+screen.LinesBelow() );
	
	if ( line < screen_lines+screen.LinesBelow() )
	{
		XXXLog("reloading 224");
		cc += linecycles;
		endcycles += linecycles;

	// update screen and border
		screen.UpdateBorder(line,border);
		if ( (Long)line < screen_lines ) 
			screen.UpdateScreen ( line, videopage + 32*n2n[line],videopage +24*8*32 + 32*(line/8) );
		if (cc<=0) goto a;
	}

	else  if ( line == screen_lines+screen.LinesBelow() )
	{
		XXXLog("reloading rest");
		cc 	  += framecycles - linecycles*(line+topborderlines);
		endcycles += framecycles - linecycles*(line+topborderlines);
	}
	// else do not reload: force z80() to exit

	XXXLogExit();

	return cc;
}


void zxsp::Run ( )
{	int n;

	int syncedframes=0;
	#if XLOG
	#define SYNCED	syncedframes++;
	#else
	#define SYNCED	
	#endif

	XLogLine ( ">zxsp::Run()" );

	XXTRAP(SCR_ABOVE>max_lines_above);
	XXTRAP(SCR_BELOW>max_lines_below);

	while (!done)
	{
		n = 0;	// n = reload T cycles
		
		if (cpu.Cycles()<=0)
		{
			cpu.Irpt();
			scanline   = -screen.LinesAbove();	// first visible scanline (negative number)
			n	   = topborderlines +scanline;	// number of invisible scanlines above
			n	  *= linecycles;		// T cycles to run without updating screen or border
		}
		
		endcycles += n;
		n = cpu.Run ( n );

		switch ( n )	// n = z80 result code
		{
		case halt_instr:
			if ( cpu.IFF1()!=enabled)
			{
				Log ( "HALT instruction executed with interrupts disabled");
				Log ( "trying an NMI to unlock the deadlock");
				cpu.NMI();
			}
			cpu.Cycles()=0;
			while (Cycles(0)){}	// wait for frame to end
			// goto tcycles_expired

		case tcycles_expired:		// frame completed

			n = soundchip.DspState();	// n = dsp state
			
			switch ( n )
			{
			case dsp_full:
				XLog("*FULL*"); 
			case dsp_high:	// on & wait
				XXLog("WAIT"); 
				screen.Sync();
				SYNCED;
				while (soundchip.DspState()>=dsp_high) { pause(); PollEvents(); }
				soundchip.Play(Now());
				if ((frameCounter-nextframe)>0) nextframe = frameCounter;	// reduce frames to do
				break;
		//	case dsp_fine:	// on				
		//		soundchip.Play(Now());
			default:	// off
		//		XXLog("FINE"); 				
				screen.Sync();
				SYNCED;
				break;
			case dsp_empty:	// on & hurry up
				XLog("*EMPTY*"); 
			case dsp_low:
				XXLog("HURRY"); 
			case dsp_fine:	// on				
				soundchip.Play(Now());
				if ((frameCounter<nextframe)<0) nextframe = frameCounter;	// increase frames to do
				break;
			}

			PollEvents();				// poll events at least once !
			while ( (frameCounter<nextframe)<0 ) 	// wait for frame interrupt
			{ pause(); PollEvents(); }

			nextframe++;
			break;
		
		case trap_instr:
			RST0();
			break;

		case irpt_error:
			Log ( "Internal error: Interrupt failure" );
			Reset();
			break;

		case nimp_instr:
			Log ( "Not implemented instruction");
			Reset();
			break;

		default:
			Log ( "Internal error: Z80 bad result: %i", n );
			Reset();
			break;
		}
	}

	XLogLine("Screen Refresh Ratio=%f%%",100.0*syncedframes/nextframe);
	XLogExit();	
}


void zxsp::RST0()
{
	XXLogLine(">zxsp::RST0()");

	if (type==zx16||type==zx48)
	{
		switch ( cpu.PC()-1 )
		{
		case 0x0556:			//  Load / Verify Tape
		//	rst0_load_verify();
		//	Cycles() = 0;
		//	frames_done=frames-1;
		//	goto x;
		case 0x04c2:			// Save Tape
		//	rst0_save();		
		//	Cycles()=0;
		//	frames_done=frames-1;
		//	goto x;
		case	0x03d5:			// BEEP
	/*		this is a very low level patch:
			the timing loop inside the beep function is patched.
			output to mic/ear is tracked by normal Output() function !!!
	
	03D5   0C			INC	C		<-- patched with RST 0
	03D6   0D	HALBZE:	DEC	C
	03D7   20FD			JR	NZ,HALBZE
	03D9   0E3F			LD	C,3FH
	03DB   05 			DEC	B
	03DC   C2D603			JP	NZ,HALBZE
	03DF
			time consumed = 1024 * B  +  16 * C  -  988  T cycles
			minus 11 T cycles which are already counted by RST 0 command in z80::Run()
	*/
			cpu.Cycles() -= 1024*(long)cpu.RB()+16*(long)cpu.RC()-999;
			cpu.BC() = 0;
			cpu.PC() = 0x03DF;
			goto x;
		}
	}

	cpu.Poke ( --cpu.SP(), cpu.PCH() );
	cpu.Poke ( --cpu.SP(), cpu.PCL() );
	cpu.PC() = 0;

x:	XXLogExit();
}


// ======================================================================

/* ----	Sound output to Mic and Tape -------------------------------------------------
*/
#define	m_ear		0x10
#define	m_mic		0x08
#define	snd_mask	(m_ear+m_mic)	// both outputs mixed to Mac speaker
#define	in_fe		0xdf		// value read from port 0xfe if line silent: 
					// 0xff (early specci rev. 2)  or  0xdf (later models)


// ======================================================================

// ----	Handler for Input instructions --------------------------------------------
Char zxsp::Input ( long cc, Short addr )
{	Char c;

	XXXLogLine ( "IN ($%04X)", (int)addr );

	switch ( (Char)addr )
	{	
	case 0x00fe:		// Keyboard
		c = in_fe;  	// bit 5 is line input *****missing*****
				// !!! => cc = endcycles - cc;
		addr = ((Short)~addr)>>8;
		for ( ptr p=keyboard.kbyte; addr; addr>>=1,p++ )
		{ if (addr&1) c&=*(Ptr)p; }
		XXXLog(" -> $%02X (%08X%08X)",(int)c, *(long*)keyboard.kbyte, *(long*)(keyboard.kbyte+4) );
		return c;

	case 0x001f:		// Kempston interface
		if (keyboard.jtype==kbdKempston)  return keyboard.jbyte;
		return 0xFF;

	case 0x00ff:		// ULA won't stop -> read mixture of 0xff (idle bus) and screen byte / attr byte
		if (scanline<topborderlines) return 0xff;		// above screen
		if (scanline>=screen_lines+topborderlines) return 0xff;	// below screen
		return videopage[(scanline-topborderlines+1)*32-(cc%linecycles/32)];	// return screen byte
		
	case 0x00fd:		// ---- zx128: paging, sound chip
		switch (addr)
		{
		case 0xfffd:	// ---- zx128: sound chip: read data register
			return soundchip.ReadRegister();
		};
	};

	if (!XXXLOG) XXLogLine ( "IN ($%04X)", (int)addr );

	return 0xFF;
}


// ----	Handler for output instructions -------------------------------------------
void zxsp::Output ( long cc, Short addr, Char n )
{
	static Char oldsound = 0x00;	

	cc = endcycles - cc;
	
	switch ( (Char)addr )
	{
	case 0x00fe:	// ---- border, mic & ear sound output
		border = n&0x07;

	// sound output
		n &= snd_mask;
		if ( n!=oldsound )
		{	oldsound = n;
			soundchip.WriteBeeper ( cc, n );
		};
		return;

	case 0x00fd:		// ---- zx128: paging, sound chip
		switch (addr)
		{
		case 0x7ffd:	PageCtrl(n);			return;
		case 0xfffd:	soundchip.SelectRegister(n);	return;
		case 0xbffd:	soundchip.WriteRegister(cc,n);	return;
		};
	};

	XXLogLine ( "OUT ($%04X,%i)", (int)addr, (int)n );
}


// ----	debugger callback stubs -------------------------
void zxsp::Info_Misc ( long, Short )	 { }
void zxsp::Info_Illegal ( long, Short )	 { }
void zxsp::Info_Weird ( long, Short )	 { }
void zxsp::Info_NMI ( long, Short )	 	 { }
void zxsp::Info_Irpt ( long, Short )	 { }
void zxsp::Info_1st_Loc ( long, Short )	 { }
void zxsp::Info_1st_Instr ( long, Short ){ }







